import errno, struct, threading, uuid
from ... import log
from .. driver_base import DriverBase
from . import websocket

ADDRESS_IN_USE_ERROR = """

Port {0} on your machine is already in use.
Perhaps BiblioPixel is already running on your machine?
"""


class SimPixel(DriverBase):

    def __init__(self, num, port=1337, pixel_positions=None, **kwds):
        """
        Args:
            num:  number of LEDs being visualizer.
            port:  the port on which the SimPixel server is running.
            pixel_positions:  the positions of the LEDs in 3-d space.
            **kwds:  keywords passed to DriverBase.
        """
        super().__init__(num, **kwds)
        self.port = port
        self.pixel_positions = self.server = self.thread = None
        self.websocks = {}

        if pixel_positions:
            self.set_pixel_positions(pixel_positions)

    def __start_server(self):
        log.debug('Starting server...')
        desc = dict(driver=self,
                    pixel_positions=self.pixel_positions,
                    selectInterval=0.001)
        try:
            self.server = websocket.Server(
                '', self.port, websocket.Client, **desc)
        except OSError as e:
            if e.errno == errno.EADDRINUSE:
                e.strerror += ADDRESS_IN_USE_ERROR.format(self.port)
                e.args = (e.errno, e.strerror)
            raise

        self.thread = threading.Thread(target=self._serve_forever, daemon=True)
        self.thread.start()

    def _serve_forever(self):
        try:
            self.server.serveforever()
        except:
            pass
        log.debug('WebSocket Server closed')

    def set_pixel_positions(self, pixel_positions):
        if not (self.thread and self.thread.is_alive()):
            # Flatten list of led positions.
            pl = [c for p in pixel_positions for c in p]
            self.pixel_positions = bytearray(struct.pack('<%sh' % len(pl), *pl))
            self.__start_server()

    def add_websock(self, oid, send_pixels):
        self.websocks[oid] = send_pixels

    def remove_websock(self, oid):
        try:
            del self.websocks[oid]
        except KeyError:
            pass

    def cleanup(self):
        self.server.close()

    def _compute_packet(self):
        self._render()

    def _send_packet(self):
        for ws in self.websocks.values():
            ws(self._buf)


# This is DEPRECATED.
DriverSimPixel = SimPixel



import gitty

HELP = """
Clear the git repository library cache.
"""


def run(args, settings):
    gitty.clear_library_cache()


def set_parser(parser):
    parser.set_defaults(run=run)
    parser.description = HELP



from .. channel_order import ChannelOrder
from .. driver_base import DriverBase
from ... util.enum import resolve_enum
from . import interfaces
from enum import IntEnum
# SubDriver imports
from . APA102 import APA102
from . LPD8806 import LPD8806
from . WS281X import WS281X
from . WS2801 import WS2801


class SPI_LEDTYPE(IntEnum):
    APA102 = 0
    DOTSTAR = 0
    WS2801 = 1
    WS281X = 2
    WS2812 = 2
    WS2812B = 2
    NEOPIXEL = 2
    LPD8806 = 3


SPI_DRIVERS = [
    APA102,
    WS2801,
    WS281X,
    LPD8806
]


def SPI(type, num, **kwargs):
    """Wrapper function for using SPI device drivers on systems like the Raspberry Pi and BeagleBone"""
    ledtype = resolve_enum(SPI_LEDTYPE, type)
    if ledtype >= len(SPI_DRIVERS):
        raise ValueError('{} is not a valid LED type.'.format(type))

    return SPI_DRIVERS[ledtype](num, **kwargs)



import copy

from . aliases import resolve_aliases

DEFAULTS = {
    'driver': {
        'typename': 'bibliopixel.drivers.SimPixel.SimPixel',
        'num': 1024
    },

    'layout': {
        'typename': 'bibliopixel.layout.matrix.Matrix',
        'width': 32,
        'height': 32,
    },

    'animation': {
        'typename': 'BiblioPixelAnimations.matrix.bloom.Bloom'
    },

    'maker': {},
    'run': {},
}

SECTIONS = tuple(sorted(DEFAULTS.keys()))


def merge(default, *others):
    result = copy.deepcopy(default)
    for o in others:
        for k, v in o.items():
            result[k].update(v)

    return result


def apply_defaults(desc):
    return merge(DEFAULTS, resolve_aliases(desc))



import threading, time
from .. import log


class AnimationThreading(object):
    """
    AnimationThreading handles threading - and eventually multiprocessing - for
    BaseAnimation.
    """

    def __init__(self, runner, run):
        self.runner = runner
        self.run = run
        self.stop_event = threading.Event()
        self.thread = None
        self.frame_overrun = False

    def stop_thread(self, wait=False):
        if self.thread:
            self.stop_event.set()
            if wait:
                self.thread.join()

    def stopped(self):
        return not (self.thread and self.thread.isAlive())

    def target(self):
        # TODO: no testpath exercises this code...
        log.debug('Starting thread...')
        self.run()
        log.debug('Thread Complete')

    def wait(self, wait_time, timestamps):
        if not wait_time:
            return

        elapsed_time = timestamps[-1] - timestamps[0]
        if elapsed_time > wait_time:
            logger = log.debug if self.frame_overrun else log.warning
            logger('Frame-time of %dms set, but took %dms!',
                   1000 * wait_time, 1000 * elapsed_time)
            self.frame_overrun = True

        elif self.runner.threaded:
            self.stop_event.wait(wait_time - elapsed_time)

        else:
            time.sleep(wait_time - elapsed_time)

    def start(self):
        if not self.runner.threaded:
            self.run()
            return

        self.stop_event.clear()

        self.thread = threading.Thread(target=self.target, daemon=True)
        self.thread.start()

        if self.runner.join_thread:
            # TODO: why would you do this rather than disable threading?
            self.thread.join()



import gitty, json, os, sys
from . import defaults
from . importer import make_object
from .. animation import runner
from .. import data_maker
from .. layout.geometry import gen_matrix
from .. layout.multimap import MultiMapBuilder
from .. util import files
from .. drivers.serial import codes


def make_layout(driver, layout, maker=None):
    maker = data_maker.Maker(**(maker or {}))
    drivers = []

    def multi_drivers(device_ids, width, height, serpentine=False, **kwds):
        build = MultiMapBuilder()

        for id in device_ids:
            build.addRow(gen_matrix(width, height, serpentine=serpentine))
            d = make_object(width=width, height=height, deviceID=id, **kwds)
            drivers.append(d)

        return build.map

    def make_drivers(multimap=False, **kwds):
        if multimap:
            return multi_drivers(**kwds)

        drivers.append(make_object(**kwds))

    coordMap = make_drivers(maker=maker, **driver)
    return make_object(drivers, coordMap=coordMap, maker=maker, **layout)


def make_animation(layout, animation, run=None):
    animation = make_object(layout, **animation)
    animation.set_runner(runner.Runner(**(run or {})))
    return animation


def project_to_animation(*, path=None, **project):
    gitty.sys_path.extend(path)

    kwds = defaults.apply_defaults(project)
    animation = kwds.pop('animation', {})
    run = kwds.pop('run', {})
    layout = make_layout(**kwds)
    return make_animation(layout, animation, run)


def apply_defaults(project, defaults):
    # TODO: consolidate with bp.project.defaults.
    for k, v in defaults.items():
        if k == 'ledtype' or not v:
            continue

        if k not in project:
            project[k] = v

        else:
            existing = project[k]
            if 'typename' not in existing:
                try:
                    existing['typename'] = v
                except:
                    pass

    ledtype = defaults.get('ledtype')
    if ledtype:
        if 'layout' not in project:
            project['layout'] = {}

        if 'type' not in project['layout']:
            project['layout']['type'] = ledtype
            # The field in LAYOUT should probably be renamed "ledtype" to be
            # consistent across the whole system.


def project_to_runnable(project, defaults=None):
    apply_defaults(project, defaults or {})
    return project_to_animation(**project).start


def run(s, is_filename=True, defaults=None):
    project = files.read_json(s, is_filename)
    runnable = project_to_runnable(project, defaults)
    runnable()



import contextlib, gitty, json, os.path, shutil


def opener(fname, mode='r', *args, **kwds):
    requests = None
    if ':' in fname:
        try:
            import requests
        except:
            pass

    if not requests:
        return open(os.path.expanduser(fname), mode, *args, **kwds)

    assert not (set(mode) & set('wax')), 'Cannot write to remote file ' + fname
    url = gitty.raw.raw(fname)
    response = requests.get(url)
    if response.ok:
        response.read = lambda: response.text
        return response

    raise IOError('%s: Can\'t open URL %s' % (response.status_code, url))


def read_json(s, is_filename=True):
    data = opener(s).read() if is_filename else s

    try:
        return json.loads(data)
    except ValueError as e:
        e.args = ['%s\nin filename %s' % (e.args[0], s)]
        raise


@contextlib.contextmanager
def remove_on_failure(dirname, remove=True):
    """Tries to create and fill a directory, removes it on failure."""
    os.makedirs(dirname)
    try:
        yield
    except:
        if remove:
            shutil.rmtree(dirname)
        raise



class MultiMapBuilder(object):

    def __init__(self):
        self.map = []
        self.offset = 0

    def addRow(self, *maps):
        yOff = len(self.map)
        lengths = [len(m) for m in maps]
        h = max(lengths)
        if(min(lengths) != h):
            raise ValueError("All maps in row must be the same height!")

        offsets = [0 + self.offset]
        count = 0
        for m in maps:
            offsets.append(h * len(m[0]) + offsets[count])
            count += 1

        for y in range(h):
            self.map.append([])
            for x in range(len(maps)):
                self.map[y + yOff] += [i + offsets[x] for i in maps[x][y]]

        self.offset = offsets[len(offsets) - 1]



from ..util import importer
from distutils.version import LooseVersion

MINIMUM_VERSIONS = {'serial': '2.7'}

INSTALL_NAMES = {
    'BiblioPixelAnimations': 'BiblioPixelAnimations',
    'noise': 'pynoise',
    'serial': 'pyserial',
}

VERSION_MESSAGE = """
You have version %s of module '%s' but you need version %s.

Please upgrade at the command line with:

    $ pip install %s --upgrade

"""

MISSING_MESSAGE = """
You are missing module '%s'.

Please install it at the command line with:

    $ pip install %s

"""


def validate_typename(typename):
    root_module = typename.split('.')[0]
    min_version = MINIMUM_VERSIONS.get(root_module)
    if not min_version:
        return

    version = __import__(root_module).VERSION
    if LooseVersion(version) >= LooseVersion(min_version):
        return

    install_name = INSTALL_NAMES.get(root_module, root_module)
    raise ValueError(VERSION_MESSAGE % (
        root_module, version, min_version, install_name))


def import_symbol(typename, package=None):
    try:
        result = importer.import_symbol(typename, package)
        package or validate_typename(typename)
        return result
    except ImportError as e:
        root_module = typename.split('.')[0]
        install_name = INSTALL_NAMES.get(root_module)
        if install_name:
            try:
                __import__(root_module)
            except ImportError:
                msg = MISSING_MESSAGE % (root_module, install_name)
                e.msg = msg + e.msg
        raise


def make_object(*args, typename, package=None, **kwds):
    """Make an object from a symbol."""
    return import_symbol(typename, package)(*args, **kwds)



from . rotation import Rotation, rotate_and_flip
from . circle import (calc_ring_pixel_count, calc_ring_steps, gen_circle,
                      pixel_positions_from_rings)
from . cube import gen_cube, pixel_positions_from_cube
from . matrix import gen_matrix, pixel_positions_from_matrix
from . strip import gen_strip



from enum import IntEnum


class CMDTYPE(IntEnum):
    SETUP_DATA = 1  # config data (LED type, SPI speed, num LEDs)
    PIXEL_DATA = 2  # raw pixel data will be sent as [R1,G1,B1,R2,G2,B2,...]
    BRIGHTNESS = 3  # data will be single 0-255 brightness value, length must be 0x00,0x01
    GETID = 4
    SETID = 5
    GETVER = 6
    SYNC = 7


class LEDTYPE(IntEnum):
    GENERIC = 0  # Use if the serial device only supports one chipset
    LPD8806 = 1
    WS2801 = 2
    # These are all the same
    WS2811 = 3
    WS2812 = 3
    WS2812B = 3
    NEOPIXEL = 3
    APA104 = 3
    # 400khz variant of above
    WS2811_400 = 4

    TM1809 = 5
    TM1804 = 5
    TM1803 = 6
    UCS1903 = 7
    SM16716 = 8
    APA102 = 9
    LPD1886 = 10
    P9813 = 11


SPIChipsets = [
    LEDTYPE.LPD8806,
    LEDTYPE.WS2801,
    LEDTYPE.SM16716,
    LEDTYPE.APA102,
    LEDTYPE.P9813
]

# Chipsets here require extra pixels padded at the end
# Key must be an LEDTYPE
# value a lambda function to calc the value based on numLEDs
BufferChipsets = {
    LEDTYPE.APA102: lambda num: (int(num / 64.0) + 1)
}



from . animation import BaseAnimation


class OffAnim(BaseAnimation):
    """A trivial animation that turns all pixels in a layout off."""

    def __init__(self, layout, timeout=10):
        super().__init__(layout)
        self.internal_delay = timeout

    def step(self, amt=1):
        self.layout.all_off()



from ... import gamma
from .. channel_order import ChannelOrder
from . base import SPIBase


class APA102(SPIBase):
    """Driver for APA102 based LED strips on devices like the Raspberry Pi and BeagleBone"""

    def __init__(self, num, gamma=gamma.APA102, **kwargs):
        super().__init__(num, gamma=gamma, **kwargs)

        # APA102 requires latch bytes at the end
        self._latch_bytes = [0xFF, 0, 0, 0] * ((num // 64) + 1)

    def _compute_packet(self):
        self._render()
        # TODO: Do all this surgery in place instead
        newBuf = [0xFF] * (self.bufByteCount() + self.numLEDs)
        newBuf[1::4] = self._buf[0::3]
        newBuf[2::4] = self._buf[1::3]
        newBuf[3::4] = self._buf[2::3]
        newBuf = [0, 0, 0, 0] + newBuf
        newBuf.extend(self._latch_bytes)
        self._packet = newBuf



from .. channel_order import ChannelOrder
from .. driver_base import DriverBase
from ... import gamma as _gamma
from ... util.enum import resolve_enum
from . import interfaces


class SPIBase(DriverBase):
    """Base driver for controling SPI devices on systems like the Raspberry Pi and BeagleBone"""
    def __init__(self, num, dev='/dev/spidev0.0',
                 interface='FILE', spi_speed=1,
                 **kwargs):

        super().__init__(num, **kwargs)

        iface = resolve_enum(interfaces.SPI_INTERFACES, interface)
        if iface >= len(interfaces._SPI_INTERFACES):
            raise ValueError('{} is not a valid interface.'.format(interface))

        self._interface = interfaces._SPI_INTERFACES[iface](dev=dev, spi_speed=spi_speed)

    def _send_packet(self):
        self._interface.send_packet(self._packet)

    def _compute_packet(self):
        self._render()
        self._packet = self._interface.compute_packet(self._buf)



from . layout import Circle, Matrix, POV, Strip, Cube
from . layout.geometry import Rotation
from . layout.geometry.matrix import gen_matrix
from . layout.geometry.circle import gen_circle
from . layout.geometry.cube import gen_cube
from . import animation, colors, font, gamma, layout, log, util

# These are DEPRECATED
from . layout import LEDCircle, LEDMatrix, LEDPOV, LEDStrip, LEDCube


def _get_version():
    from os.path import abspath, dirname, join
    filename = join(dirname(abspath(__file__)), 'VERSION')
    return open(filename).read().strip()


__version__ = _get_version()
VERSION = __version__
